import click
import textwrap

from catacomb.common import constants, errors

from terminaltables import AsciiTable


def color_text(text, color):
    """Adds color to the provided text.

    Arguments:
        text (str): The text to color.
        color (str): The color as a string, e.g. 'red', 'green' and 'blue'
            (more: http://click.pocoo.org/6/api/#click.style).

    Returns:
        A `string` with ANSI color codes appended to it, or just the regular
        non-colored text if the color is not available.
    """
    try:
        return click.style(text, fg=color)
    except TypeError:
        # Invalid color specified.
        print_error(errors.INVALID_COLOR.format(color))
    return text


def to_table(headers, rows):
    """Produces a nicely formatted ASCII table from the provided data.

    Arguments:
        headers (list): The headers of the table.
        rows (list): List of rows to append to the table.

    Returns:
        A `string` representation of the table.
    """
    table = [headers]
    table.extend(rows)
    return AsciiTable(table).table


def create_row(alias, command, description):
    """Creats a new table row using the provided attributes for a specific
    command.

    Arguments:
        alias (str): An alias that the command is referenced by.
        command (str): The command that is stored.
        description (str): A description of the command stored.

    Returns:
        A `list` representing the new row for a command.
    """
    return [
        textwrap.fill(alias, constants.MAX_TABLE_WIDTH),
        textwrap.fill(command, constants.MAX_TABLE_WIDTH),
        textwrap.fill(description, constants.MAX_TABLE_WIDTH),
    ]


def print_error(message):
    """Correctly styles and prints an error message to standard output.

    Arguments:
        message (str): An error message.
    """
    click.echo(color_text(message, "red"), err=True)


def print_success(message):
    """Correctly styles and prints a success message to standard output.

    Arguments:
        message (str): A success message.
    """
    click.echo(color_text(message, "green"))



import click

from catacomb.common import constants, errors
from catacomb.utils import tomb_handler, formatter


@click.command(
    constants.CMD_RM_NAME, help=constants.CMD_RM_DESC,
    short_help=constants.CMD_RM_DESC)
@click.argument("alias", nargs=1)
@click.pass_context
def rm(ctx, alias):
    """Removes a command from the tomb.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        alias (str): The alias of the command to remove.
    """
    if tomb_handler.remove_command(ctx, alias):
        formatter.print_success(constants.CMD_RM_OK.format(alias))
    else:
        formatter.print_error(errors.ALIAS_NOT_FOUND.format(alias))



import click
import os


class PluginLoader(click.MultiCommand):
    """The `PluginLoader` handles the loading of plugins from the specified
    directory. This implementation supports lazy loading of commands from
    plugins that can then be injected straight into a group.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set the path to the directory that we're loading plugins from. The
        # subdirectory is passed through the kwargs argument and is derived
        # from the name of the function calling this loader, e.g. calling from
        # function `tomb_entry` will load plugins from `commands/tomb`.
        subdir = kwargs["name"][:-6]
        self._plugin_dir = os.path.join(
            os.path.dirname(__file__), os.pardir, "commands", subdir)

    def list_commands(self, ctx):
        """Returns a list of subcommands that are present in the plugin
        directory.

        Arguments:
            ctx (click.Context): Holds the state relevant for script execution.

        Returns:
            A sorted list of subcommands.
        """
        commands = []
        for filename in os.listdir(self._plugin_dir):
            if filename.endswith(".py") and "__" not in filename:
                commands.append(filename[:-3])
        commands.sort()
        return commands

    def get_command(self, ctx, name):
        """Given a context and a command name, this returns a Command object
        if it exists or returns None.

        Arguments:
            ctx (click.Context): Holds the state relevant for script execution.
            name (str): The name of command.

        Returns:
            A `click.Command` object if the plugin exists, otherwise None.
        """
        namespace = {}
        plugin = os.path.join(self._plugin_dir, name + ".py")
        with open(plugin) as f:
            code = compile(f.read(), plugin, "exec")
            eval(code, namespace, namespace)
        return namespace[name]



import click

from catacomb.common import constants, errors
from catacomb.utils import tomb_handler, formatter


@click.command(
    constants.CMD_ADD_NAME, help=constants.CMD_ADD_DESC,
    short_help=constants.CMD_ADD_DESC)
@click.argument("alias", nargs=1)
@click.argument("command", nargs=-1, required=False)
@click.pass_context
def add(ctx, alias, command):
    """Adds a new command to the tomb.

    First prompting the user to specify the desired alias as well as a
    description of what the command does.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        alias (str): The alias of the command being added to the tomb.
        command (str): The command to add to the tomb.
    """
    if tomb_handler.get_command(ctx, alias):
        # If the `alias` specified is already used for a command in the tomb,
        # prompt the user to determine if they want to overwrite the old
        # command.
        update = click.prompt(constants.PROMPT_UPDATE.format(alias))
        if update.lower() != "y":
            # Abort the action.
            formatter.print_error(errors.ACTION_ABORTED)
            return

    if not command:
        # If the `command` hasn't yet been specified we'll need to prompt the
        # user for that as well.
        command = click.prompt(constants.PROMPT_CMD)
    else:
        # The argument `command` is passed as a tuple, so we'll need to convert
        # it to a string in order to evaluate it later.
        command = " ".join(command)
    description = click.prompt(constants.PROMPT_DESCR)

    # Save the new command to the current tomb.
    tomb_handler.add_command(ctx, command, alias, description)
    formatter.print_success(constants.CMD_ADD_OK.format(alias, description))



import click
import os

from catacomb.common import constants, errors
from catacomb.utils import tomb_handler, formatter


@click.command(
    constants.CMD_USE_NAME, help=constants.CMD_USE_DESC,
    short_help=constants.CMD_USE_DESC)
@click.argument("alias", nargs=1)
@click.pass_context
def use(ctx, alias):
    """Retrieves a command from the tomb and executes it.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        alias (str): The alias of the command to execute.
    """
    cmd = tomb_handler.get_command(ctx, alias)

    if cmd:
        # Execute the retrieved command.
        os.system(cmd)
    else:
        # The command alias doesn't exist.
        formatter.print_error(errors.ALIAS_NOT_FOUND.format(alias))



import click

from catacomb.common import constants, errors
from catacomb.utils import tomb_handler, formatter


@click.command(
    constants.CMD_LIST_NAME, help=constants.CMD_LIST_DESC,
    short_help=constants.CMD_LIST_DESC)
@click.pass_context
def list(ctx):
    """Lists all the commands in the current tomb.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
    """
    table = tomb_handler.tomb_to_table(ctx)
    if table is None:
        formatter.print_error(errors.EMPTY_TOMB)
    else:
        click.echo(table)



# User prompts.
PROMPT_CMD = "Command"
PROMPT_DESCR = "Description"
PROMPT_CLEAN = (
    "You're about to completely destroy the contents of this "
    "tomb. Would you like to continue? (y/n)")
PROMPT_UPDATE = (
    "The alias '{0}' is already being used in this tomb. Would "
    "you like to overwrite it? (y/n)")

# Table properties.
TABLE_HEADERS = ["Alias", "Command", "Description"]

# Limits.
MAX_TABLE_WIDTH = 30

# Command definitions.
CMD_ADD_NAME = "add"
CMD_ADD_DESC = "Stores a new command in the tomb."
CMD_ADD_OK = (
    "Successfully added the command with alias '{0}', and "
    "description '{1}' to the tomb.")

CMD_CLEAN_NAME = "clean"
CMD_CLEAN_DESC = "Empties the contents of a tomb."
CMD_CLEAN_OK = "The contents of the tomb have been cleared."
CMD_CLEAN_FORCE_DESC = "Ignore the prompt for user confirmation."

CMD_USE_NAME = "use"
CMD_USE_DESC = "Grabs a command from the tomb and executes it."

CMD_LIST_NAME = "list"
CMD_LIST_DESC = "Lists the commands currently stored in the tomb."

CMD_RM_NAME = "rm"
CMD_RM_DESC = "Removes a command from the tomb."
CMD_RM_OK = "Successfully removed the command with alias '{0}' from the tomb."



import click

from catacomb.common import constants, errors
from catacomb.utils import tomb_handler, formatter


@click.command(
    constants.CMD_CLEAN_NAME, help=constants.CMD_CLEAN_DESC,
    short_help=constants.CMD_CLEAN_DESC)
@click.option(
    "--force", "-f", is_flag=True, default=False,
    help=constants.CMD_CLEAN_FORCE_DESC)
@click.pass_context
def clean(ctx, force):
    """Prompts the user to confirm cleaning of the current tomb. If the user
    confirms the action, the tomb will be reset to it's original empty state.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        force (bool): Option to disable prompting the user for confirmation.
    """
    if not force:
        # Prompt the user for command details.
        confirm = click.prompt(constants.PROMPT_CLEAN)

    if force or confirm.lower() == "y":
        tomb_handler.clean_tomb(ctx)
        formatter.print_success(constants.CMD_CLEAN_OK)
    else:
        formatter.print_error(errors.ACTION_ABORTED)



import click

from catacomb import settings
from catacomb.common import about
from catacomb.decorators.context import Context
from catacomb.utils.plugin_loader import PluginLoader

pass_context = click.make_pass_decorator(Context, ensure=True)


@click.group(
    context_settings=settings.CONTEXT_SETTINGS, help=about.description,
    cls=PluginLoader)
@pass_context
def tomb_entry(ctx):
    """Entry point for the `tomb` command. Passes responsibility of handling
    a command down the chain of subcommand handlers until we've found one that
    can handle the given subcommand.

    This entry point handles actions involving the contents of the current
    tomb.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
    """
    pass



import json

from catacomb.common import constants
from catacomb.utils import formatter


def read_tomb(ctx):
    """Reads the contents of a tomb.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.

    Returns:
        A `dict` representing the contents of the tomb.
    """
    with open(ctx.obj.config_path, "r") as f:
        json_data = json.load(f)
    return json_data


def write_tomb(ctx, data):
    """Replaces the current contents of the tomb with `data`.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        data (dict): The data to store in the tomb.
    """
    with open(ctx.obj.config_path, "w") as f:
        f.write(json.dumps(data, indent=2))


def clean_tomb(ctx):
    """Clears the entire contents of the tomb, resetting it to it's original
    state.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
    """
    write_tomb(ctx, {})


def add_command(ctx, command, alias, description):
    """Adds a new command to the current tomb.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        command (str): The command to add.
        alias (str): The alias to save the command as.
        description (str): What the command does.
    """
    data = read_tomb(ctx)

    data[alias] = {
        "command": command,
        "description": description
    }

    write_tomb(ctx, data)


def get_command(ctx, alias):
    """Retrieves a command from the current tomb, using its alias.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        alias (str): The alias to save the command as.

    Returns:
        The command as a `string`, or None if not found.
    """
    data = read_tomb(ctx)

    if alias in data:
        return data[alias]["command"]
    return None


def remove_command(ctx, alias):
    """Removes a command from the current tomb.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.
        alias (str): The alias to save the command as.

    Returns:
        A `bool`, True if the alias could be removed, False otherwise.
    """
    data = read_tomb(ctx)

    if alias not in data:
        return False

    # Remove the command then write back to the file.
    del data[alias]
    write_tomb(ctx, data)

    return True


def tomb_to_table(ctx):
    """Converts the current tomb to a table containing information about each
    command stored.

    Arguments:
        ctx (click.Context): Holds the state relevant for script execution.

    Returns:
        A `string` representation of the table.
    """
    data = read_tomb(ctx)

    # Convert each stored command to it's own row.
    rows = []
    for alias in data.keys():
        cmd = data[alias]["command"]
        desc = data[alias]["description"]
        rows.append(formatter.create_row(alias, cmd, desc))

    if len(rows):
        return formatter.to_table(constants.TABLE_HEADERS, rows)
    return None
